<?php
/**
 * The plugin class contains convenience functions for working with the
 * installation, upgrade, and removal of plugins from Zen Cart.
 *
 * To make use of this class the following file needs to be created:
 * <strong>/admin/includes/classes/<i>unique_key</i>_plugin.php</strong>:
 * A subclass of plugin (a class extending plugin). Any of the public methods
 * in the plugin class can be overridden and some must be overridden.
 *
 * The following optional files are recommended:
 * <strong>/admin/includes/init_includes/init_<i>unique_key</i>_install.php</strong>:
 * When called starts the install (loaded by the auto_loader).
 *
 * <strong>/admin/includes/init_includes/init_<i>unique_key</i>_remove.php</strong>:
 * When called starts the removal (loaded by the auto_loader).
 *
 * <strong>/admin/includes/auto_loaders/config.<i>unique_key</i>_onetime.php</strong>:
 * A file to start one of the init includes scripts (install / remove).
 *
 * @see plugin::createInitIncludeFile()
 * @see plugin::createOnetimeFile()
 * @copyright Copyright 2012 - 2013 Andrew Ballanger
 * @license http://www.gnu.org/licenses/gpl.txt GNU GPL V3.0
 */
class plugin {

	/**
	 * Retrieves the version number of the plugin being installed, upgraded,
	 * or removed.
	 *
	 * Subclasses should override this method
	 * @return string the version
	 */
	public function getVersion() { return '0.3'; }

	/**
	 * Retrieves the unique key used as a prefix for configuration options
	 * stored in the database and as a prefix for language defines. The
	 * unique key should always start with a letter and all letters should
	 * be capital. The prefix should only include alpha numerical characters.
	 *
	 * Subclasses should override this method.
	 * @return string the unique key to use as a prefix
	 */
	public function getUniqueKey() { return 'PLUGIN'; }

	/**
	 * Retrieves the unique name displayed in the Zen Cart configuration menu
	 * and during installation, upgrade, or removal of the plugin.
	 *
	 * Subclasses should override this method.
	 * @return string the unique name for this plugin
	 */
	public function getUniqueName() { return 'plugin'; }

	/**
	 * Retrieves the description of this plugin. The description is saved in the
	 * database along with the configuration group. Currently it is not shown in
	 * the Zen Cart admin.
	 *
	 * Subclasses can optionally override this method.
	 * @return string a short description of this plugin
	 */
	public function getDescription() { return ''; }

	/**
	 * Retrieves the default configuration options for this plugin.
	 *
	 * Subclasses can optionally override this method.
	 * @return multitype:(multitype:string)
	 */
	public function getDefaultConfiguration() { return array(); }

	/**
	 * Retrieves the list of new files added by this plugin to Zen Cart.
	 * This should be an array of fully qualified file names.
	 *
	 * Subclasses should override this method.
	 * @return multitype:string
	 */
	public function getNewFiles() { return array(); }

	/**
	 * Retrieves the list of obsolete files added by previous versions of this
	 * plugin to Zen Cart. This should be an array of fully qualified file names.
	 *
	 * It is highly recommended to include "blank" versions of these files in
	 * your plugin distribution to overwrite the obsolete version and avoid
	 * potential conflicts if obsolete files are loaded by Zen Cart.
	 *
	 * Subclasses should override this method.
	 * @return multitype:string
	 */
	public function getObsoleteFiles() { return array(); }

	/**
	 * Handles any changes to files required by this plugin.
	 * This is typically used to verify required modifications to core Zen Cart
	 * files have been made or to actually make changes to core Zen Cart files.
	 *
	 * Subclasses can optionally override this method.
	 * @param boolean $install true if installing / upgrading, false otherwise
	 * @return boolean true if the changes succeed, false otherwise.
	 */
	public function handleFileChanges($install = true) { return true; }

	/**
	 * Handles any database changes required by this plugin.
	 *
	 * Subclasses can optionally override this method.
	 * @param boolean $install true if installing / upgrading, false otherwise
	 * @return boolean true if the changes succeed, false otherwise.
	 */
	public function handleDatabaseChanges($install = true) { return true; }

	/**
	 * Handles the installation and removal of any new admin menus required
	 * by this plugin.
	 *
	 * Subclasses can optionally override this method.
	 * @param boolean $install true if installing / upgrading, false otherwise
	 * @return boolean true if the changes succeed, false otherwise.
	 */
	public function handleAdminMenuChanges($install = true) { return true; }

	/**
	 * Handles any special "upgrade" changes needed from previously installed
	 * versions of the plugin.
	 *
	 * For example database changes or configuration key / value changes.
	 *
	 * Subclasses can optionally override this method.
	 * @param string $version
	 * @return boolean true if the upgrade is successful, false otherwise.
	 */
	public function handleUpgradeFrom($version) { return true; }

	/**
	 * Constructs a new plugin and loads necessary language files.
	 */
	public function __construct() {
		global $messageStack;

		// load the language files for the installer (fall back to english)
		$lang_file = zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . $_SESSION['language'] . '/modules/plugin', '/plugin.php', 'false');
		if((include_once $lang_file) === false) {
			$lang_file = zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . 'english/modules/plugin', '/plugin.php', 'false');
			if((include_once $lang_file) === false) {
				$messageStack->add('The core language file for installing, upgrading, or removing this plugin could not be loaded. Please check your upload!', 'error');
			}
		}

		// load any additional language files (fall back to english)
		$lang_file = zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . $_SESSION['language'] . '/modules/plugin', '/' . strtolower($this->getUniqueKey()) . '.php', 'false');
		if((@include_once $lang_file) === false) {
			$lang_file = zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . 'english/modules/plugin', '/' . strtolower($this->getUniqueKey()) . '.php', 'false');
			@include_once $lang_file;
		}

		// Display warning messages if anything in the admin configure.php
		// appears to be amiss (or just plain strange)
		$this->checkConfigure();
	}

	public function install() {
		global $messageStack;

		// Verify new files have been uploaded
		$success = $this->verifyNewFiles();

		// Handle any files changes
		if($success) { $success = $this->handleFileChanges(); }

		// Perform an upgrade if needed
		if($success) { $success = $this->upgrade(); };

		// Install any needed tables
		if($success) $success = $this->handleDatabaseChanges();

		// Install the configuration group
		if($success) {
			$gid = $this->getConfigurationGroupId();
			if($gid == -1) $success = false;
		}

		// Install (or update) any needed configuration items
		if($success) $success = $this->updateConfiguration();

		// Install (or update) any admin menus and items
		if($success && function_exists('zen_register_admin_page')) {
			$success = $this->handleAdminMenuChanges();
			if(!$success) $messageStack->add(sprintf(
				PLUGIN_INSTALL_ERROR_AUTOLOAD,
				$this->getUniqueName()
			), 'error');
		}

		// Install (or update) the version number
		if($success) {
			$success = $this->updateVersionNumber();
		}

		$this->statusMessage($success);
		return $success;
	}

	public function upgrade() {
		global $messageStack;

		$success = $this->removeObsoleteFiles();

		$version = $this->getInstalledVersion();
		if($version !== '-1') {
			$messageStack->add(sprintf(
				PLUGIN_INSTALL_FOUND_PREVIOUS_VERSION,
				$this->getUniqueName(),
				$version,
				$this->getVersion()
			), 'success');

			return $success && $this->handleUpgradeFrom($version);
		}

		return $success;
	}

	public function remove() {

		// Remove installed tables
		$success = $this->handleDatabaseChanges(false);

		// Remove installed admin menus and items
		if($success && function_exists('zen_register_admin_page')) {
			$success = $this->handleAdminMenuChanges(false);
			if(!$success) $messageStack->add(sprintf(
				PLUGIN_REMOVE_ERROR_AUTOLOAD,
				$this->getUniqueName()
			), 'error');
		}

		// Remove installed configuration items
		if($success) $success = $this->removeConfiguration();

		// Remove
		if($success) $success = $this->handleFileChanges(false);

		// Remove any new files which have been uploaded
		if($success) $success = $this->removeNewFiles();

		// Remove the version number
		if($success) $success = $this->removeVersionNumber();

		$this->statusMessage($success, false);

		return $success;
	}

	/**
	 * Verifies the new files required by this module are uploaded. This will
	 * automatically check for the presence of required installation and removal
	 * files.
	 *
	 * @see plugin::getNewFiles()
	 * @return boolean true if the verification passes, false otherwise.
	 */
	protected function verifyNewFiles() {
		global $messageStack;

		$success = true;

		// Verify new files
		foreach($this->getNewFiles() as $file) {
			if(!file_exists($file)) {
				$messageStack->add(sprintf(PLUGIN_INSTALL_ERROR_FILE_NOT_FOUND, $this->getUniqueName(), $file), 'error');
				$success = false;
			}
		}

		// Verify installer file(s)
		foreach($this->getAutoLoaderFiles() as $key => $file) {
			if(!file_exists($file)) {
				$messageStack->add(
					sprintf(
						PLUGIN_INSTALL_ERROR_FILE_NOT_FOUND,
						$this->getUniqueName(),
						$file
					),
					'error'
				);
				$success = false;
			}
		}

		return $success;
	}

	/**
	 * Removes the new files installed by this module.
	 *
	 * @see plugin::getNewFiles()
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function removeNewFiles() {
		global $messageStack;

		$success = true;
		foreach($this->getNewFiles() as $file) {
			@unlink($file);

			if(file_exists($file)) {
				$messageStack->add(sprintf(PLUGIN_REMOVE_ERROR_FILE_FOUND, $this->getUniqueName(), $file), 'error');
				$success = false;
			}
		}
		return $success;
	}

	/**
	 * Removes the obsolete files installed by previous versions of this module.
	 *
	 * @see plugin::getObsoleteFiles()
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function removeObsoleteFiles() {
		global $messageStack;

		$success = true;
		foreach($this->getObsoleteFiles() as $file) {
			// Remove this file if it exists (should no longer be used)
			if(file_exists($file) && !unlink($file)) {
				$messageStack->add(
					sprintf(
						PLUGIN_INSTALL_ERROR_FILE_FOUND,
						$this->getUniqueName(),
						$file
					),
					'error'
				);
				$success = false;
			}
		}
		return $success;
	}

	/**
	 * Retrieves the installed version of this plugin.
	 *
	 * By default the version is stored in the configuration table
	 * using "$this->getUniqueKey() . '_VERSION'".
	 *
	 * @return string the installed version, or '-1' if not installed.
	 */
	protected function getInstalledVersion() {
		global $db;

		$retval = '-1';

		// Grab the currently installed version
		$check = $db->Execute(
			'SELECT `configuration_value` FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key` = \'' . $this->getUniqueKey() . '_VERSION\''
		);
		if(!$check->EOF) {
			$retval = $check->fields['configuration_value'];
		}

		return $retval;
	}

	/**
	 * Removes the version number of this plugin from the database.
	 *
	 * By default the version is stored in the configuration table
	 * using "$this->getUniqueKey() . '_VERSION'".
	 *
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function removeVersionNumber() {
		global $db;

		$success = true;

		$db->Execute(
			'DELETE FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key`=\'' . $this->getUniqueKey() . '_VERSION\''
		);
		$check = $db->Execute(
			'SELECT `configuration_value` FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key` = \'' . $this->getUniqueKey() . '_VERSION\''
		);
		if(!$check->EOF) {
			$messageStack->add(
				sprintf(
					PLUGIN_REMOVE_ERROR_DATABASE_TABLE,
					$this->getUniqueName(),
					TABLE_CONFIGURATION,
					'configuration_key: ' . $key
				),
				'error'
			);
			$success = false;
		}

		return $success;
	}

	/**
	 * Retrieves the configuration group id for this plugin based upon
	 * the unique name of the plugin.
	 *
	 * If the configuration group id cannot be found in the database, the
	 * configuration group id will be added to the database.
	 *
	 * @return integer the configuration group id, or -1 if it cannot be added
	 */
	protected function getConfigurationGroupId() {
		global $db, $messageStack;

		$retval = -1;

		if($this->getUniqueName() === null) return $retval;

		$check = $db->Execute(
			'SELECT `configuration_group_id` FROM `' . TABLE_CONFIGURATION_GROUP . '` ' .
			'WHERE `configuration_group_title` = \'' . $this->getUniqueName() . '\''
		);
		if($check->EOF) {
			$max_sort = $db->Execute(
				'SELECT MAX(sort_order) AS `max_sort` FROM `' . TABLE_CONFIGURATION_GROUP . '`'
			);
			if(!$max_sort->EOF) {
				$max_sort = $max_sort->fields['max_sort'] + 1;

				// Create configuration group
				$db->Execute(
					'INSERT INTO `' . TABLE_CONFIGURATION_GROUP . '` ' .
					'VALUES (\'\', \'' . $this->getUniqueName() . '\', ' .
					'\''. $this->getDescription() . '\', \'' . $max_sort . '\', \'1\')'
				);
				$retval = (int) $db->insert_ID();
			}
			else {
				$messageStack->add(sprintf(PLUGIN_INSTALL_ERROR_SORT_ORDER, $this->getUniqueName(), TABLE_CONFIGURATION_GROUP), 'error');
			}
		}
		else {
			$retval = (int) $check->fields['configuration_group_id'];
		}

		if($retval == -1) {
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_DATABASE_KEY,
					$this->getUniqueName(),
					TABLE_CONFIGURATION_GROUP,
					'configuration_group_title: ' . $this->getUniqueName()
				),
				'error'
			);
		}

		return $retval;
	}

	/**
	 * Updates the configuration options stored in the database for this plugin
	 * using the default configuration.
	 *
	 * This will update all portions of an option except the "key" and "value".
	 * If the configuration option does not exist it will be added with the
	 * specified default value.
	 *
	 * @see plugin::getDefaultConfiguration()
	 * @return boolean true if the update is successful, false otherwise.
	 */
	protected function updateConfiguration() {
		global $db;

		// Some default data used for configuration entries
		$sql_data_array = array(
			'configuration_group_id' => $this->getConfigurationGroupId(),
			'sort_order' => 1, // Start at 1, version is at 0
			'last_modified' => 'null',
			'date_added' => 'now()'
		);

		foreach($this->getDefaultConfiguration() as $key => $data) {
			// Add the unique key as a prefix
			$key = $this->getUniqueKey() . '_' . $key;

			// Checks presence in database (do not trust it to be defined)
			$check = $db->Execute(
				'SELECT `configuration_id` FROM `' . TABLE_CONFIGURATION . '` ' .
				'WHERE `configuration_key`=\'' . $key . '\''
			);
			if($check->EOF) {
				$data['configuration_key'] = $key;
				$data['configuration_title'] = constant($key . '_TITLE');
				$data['configuration_description'] = constant($key . '_DESCRIPTION');
				zen_db_perform(TABLE_CONFIGURATION, array_merge($sql_data_array, $data));
			}
			else {
				unset($data['configuration_value']);
				$data['sort_order'] = $sql_data_array['sort_order'];
				$this->updateConfigurationOption($key, $data);
			}

			$sql_data_array['sort_order']++;
		}

		return true;
	}

	/**
	 * Removes the configuration entries from the database.
	 *
	 * @see plugin::getDefaultConfiguration()
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function removeConfiguration() {
		global $db, $messageStack;

		$success = true;

		// Remove default configuration items
		foreach($this->getDefaultConfiguration() as $key => $data) {
			$key = $this->getUniqueKey() . '_' . $key;
			if(!removeConfigurationOption($key)) {
				$messageStack->add(
					sprintf(
						PLUGIN_REMOVE_ERROR_DATABASE_KEY,
						$this->getUniqueName(),
						TABLE_CONFIGURATION,
						'configuration_key: ' . $key
					),
					'error'
				);
				$success = false;
			}
		}

		// Remove the configuration group id
		$db->Execute(
			'DELETE FROM `' . TABLE_CONFIGURATION_GROUP . '` ' .
			'WHERE `configuration_group_title`=\'' . $this->getUniqueName() . '\''
		);
		$check = $db->Execute(
			'SELECT `configuration_group_id` FROM `' . TABLE_CONFIGURATION_GROUP . '` ' .
			'WHERE `configuration_group_title`=\'' . $this->getUniqueName() . '\''
		);
		if(!$check->EOF) {
			$messageStack->add(
				sprintf(
					PLUGIN_REMOVE_ERROR_DATABASE_KEY,
					$this->getUniqueName(),
					TABLE_CONFIGURATION_GROUP,
					'configuration_group_title: ' . $this->getUniqueName()
				),
				'error'
			);
			$success = false;
		}

		return $success;
	}

	/**
	 * Updates the version number of the plugin in the database.
	 * If not already present the version will be added.
	 *
	 * By default the version is stored in the configuration table
	 * using "$this->getUniqueKey() . '_VERSION'".
	 *
	 * @return boolean true if the update is successful, false otherwise.
	 */
	protected function updateVersionNumber() {
		global $db;

		$success = true;

		$key = $this->getUniqueKey() . '_VERSION';

		// Some default data used for configuration entries
		$sql_data_array = array(
			'configuration_group_id' => $this->getConfigurationGroupId(),
			'sort_order' => 0,
			'configuration_key' => $key,
			'configuration_title' => sprintf(PLUGIN_OPTION_VERSION_TITLE, $this->getUniqueName()),
			'configuration_description' => sprintf(PLUGIN_OPTION_VERSION_DESCRIPTION, $this->getUniqueName()),
			'configuration_value' => $this->getVersion(),
			'set_function' => 'zen_cfg_select_option(array(\\\'' . $this->getVersion() . '\\\'),'
		);

		$check = $db->Execute(
			'SELECT `configuration_id` FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key` = \'' . $key . '\''
		);
		if($check->EOF) {
			$sql_data_array['date_added'] = 'now()';
			zen_db_perform(TABLE_CONFIGURATION, $sql_data_array);
		}
		else {
			$sql_data_array['last_modified'] = 'now()';
			zen_db_perform(
				TABLE_CONFIGURATION, $sql_data_array, 'update',
				'`configuration_id`=\'' . $check->fields['configuration_id'] . '\''
			);
		}
		unset($sql_data_array);

		$check = $db->Execute(
			'SELECT `configuration_id` FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key` = \'' . $key . '\''
		);
		if($check->EOF) {
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_DATABASE_KEY,
					$this->getUniqueName(),
					TABLE_CONFIGURATION,
					'configuration_key: ' . $key
				),
				'error'
			);
			$success = false;
		}

		return $success;
	}

	/**
	 * Displays a status message (success / failure) when the script ends.
	 * If successful this also disables the script from running again.
	 *
	 * @param boolean $success the current status of the script
	 * @param boolean $install true if installing / upgrading, false otherwise
	 */
	protected function statusMessage($success = true, $install = true) {
		global $messageStack;

		$formatPrefix = 'PLUGIN_' . ($install ? 'INSTALL' : 'REMOVE') . '_';
		if($success) {
			$files = $this->getAutoLoaderFiles();

			// Verify we have permission to modify the files
			foreach($files as $key => $file) {
				if(!is_writable($file)) {
					$messageStack->add(
						sprintf(
							constant($formatPrefix . 'ERROR_FILE_WRITE'),
							$this->getUniqueName(),
							$file
						),
						'error'
					);
					$success = false;
				}
			}

			// Try to remove the files
			if($success) {
				if($install) {
					if(!@unlink($files['loader'])) {
						$messageStack->add(
							sprintf(
								constant($formatPrefix . 'ERROR_AUTOLOAD'),
								$this->getUniqueName(),
								$files['loader']
							),
							'error'
						);
						$success = false;
					}
				}
				else {
					foreach($files as $key => $file) {
						if(!@unlink($file)) {
							$messageStack->add(
								sprintf(
									constant($formatPrefix . 'ERROR_AUTOLOAD'),
									$this->getUniqueName(),
									$file
								),
								'error'
							);
							$success = false;
						}
					}

					// Remove any plugin specific installer language files.
					// (these may not be present so ignore errors)
					@unlink(zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . $_SESSION['language'] . '/modules/plugin', '/' . strtolower($this->getUniqueKey()) . '.php', 'false'));
					@unlink(zen_get_file_directory(DIR_FS_ADMIN . DIR_WS_LANGUAGES . 'english/modules/plugin', '/' . strtolower($this->getUniqueKey()) . '.php', 'false'));
				}
			}
		}

		// Display message
		$messageStack->add(
			sprintf(
				constant($formatPrefix . ($success ? 'SUCCESS' : 'ERROR')),
				$this->getUniqueName()
			),
			($success ? 'success' : 'error')
		);

		return $success;
	}

	/**
	 * Returns a list of the files utilized by the auto loader scripts for
	 * installing, upgrading, or removing this plugin.
	 *
	 * @return multitype:string the locations of the files
	 */
	protected function getAutoLoaderFiles() {
		return array(
			'class' => DIR_FS_ADMIN . DIR_WS_CLASSES .
				strtolower($this->getUniqueKey()) . '_plugin.php',
			'loader' => DIR_FS_ADMIN . DIR_WS_INCLUDES . 'auto_loaders/config.' .
				strtolower($this->getUniqueKey()) . '_onetime.php',
			'install' => DIR_FS_ADMIN . DIR_WS_INCLUDES . 'init_includes/init_' .
				strtolower($this->getUniqueKey()) . '_install.php',
			'remove' => DIR_FS_ADMIN . DIR_WS_INCLUDES . 'init_includes/init_' .
				strtolower($this->getUniqueKey()) . '_remove.php'
		);
	}

	/**
	 * Determines if the table exists in the database.
	 *
	 * @param string $table the name of the table in the database
	 * @return boolean true if the table exists, false otherwise
	 */
	protected function dbTableExists($table) {
		global $db;

		$check = $db->Execute(
			'SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS ' .
			'WHERE TABLE_SCHEMA = \'' . DB_DATABASE . '\' ' .
			'AND TABLE_NAME = \'' . $db->prepare_input($table) . '\''
		);
		return !$check->EOF;
	}

	/**
	 * Determines if the column exists in the table in the database.
	 *
	 * @param string $table the name of the table in the database
	 * @param string $column the name of the column in the table
	 * @return boolean true if the column exists in the table, false otherwise
	 */
	protected function dbColumnExists($table, $column) {
		global $db;

		$check = $db->Execute(
			'SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS ' .
			'WHERE TABLE_SCHEMA = \'' . DB_DATABASE . '\' ' .
			'AND TABLE_NAME = \'' . $db->prepare_input($table) . '\' ' .
			'AND COLUMN_NAME = \''. $db->prepare_input($column) . '\''
		);
		return !$check->EOF;
	}

	/**
	 * Determines if the key exists in the table in the database.
	 *
	 * @param string $table the name of the table in the database
	 * @param string $key the name of the key in the table
	 * @return boolean true if the key exists in the table, false otherwise
	 */
	protected function dbKeyExists($table, $key) {
		global $db;

		$check = $db->Execute(
			'SHOW KEYS FROM `' . $db->prepare_input($table) . '` ' .
			'WHERE `key_name`=\'' . $db->prepare_input($key) . '\''
		);
		return !$check->EOF;
	}

	/**
	 * Updates the configuration option in the database using the supplied
	 * configuration data. If the configuration option does not exist this method
	 * will do nothing. A new key to use for the option can be specified in the
	 * data array (using 'configuration_key'). If not specified in the data array
	 * the title and description for the option will be updated using the defined
	 * language constants for the key.
	 *
	 * The language constants for the key are determined by taking the key and
	 * appending _TITLE and _DESCRIPTION respectively.
	 *
	 * @param string $key the configuration key to update.
	 * @param array $data the array of configuration settings.
	 */
	protected function updateConfigurationOption($key, $data = array()) {
		global $db;

		// If a new key was sent, make sure we use the new key
		$new_key = (array_key_exists('configuration_key', $data) ? $data['configuration_key'] : $key);

		$check = $db->Execute(
			'SELECT `configuration_id` FROM `' . TABLE_CONFIGURATION . '` ' .
			'WHERE `configuration_key`=\'' . $key . '\''
		);
		if(!$check->EOF) {
			$sql_data_array = array(
				'configuration_key' => $new_key,
				'configuration_title' => @constant($new_key . '_TITLE'),
				'configuration_description' => @constant($new_key . '_DESCRIPTION'),
				'last_modified' => 'now()'
			);

			zen_db_perform(TABLE_CONFIGURATION, array_merge($sql_data_array, $data), 'update', '`configuration_id`=\'' . $check->fields['configuration_id'] .'\'');
		}
	}

	/**
	 * Removes a specific configuration option if present.
	 *
	 * @param string $key the configuration key to update.
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function removeConfigurationOption($key) {
		global $db;

		if(zen_not_null($key) && defined($key)) {
			$db->Execute(
				'DELETE FROM `' . TABLE_CONFIGURATION . '` ' .
				'WHERE `configuration_key`=\'' . $key . '\''
			);
			$check = $db->Execute(
				'SELECT `configuration_id` FROM `' . TABLE_CONFIGURATION . '` ' .
				'WHERE `configuration_key` = \'' . $key . '\''
			);
			if(!$check->EOF) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Checks the configure.php files for some common errors which can adversely
	 * affect Zen Cart and modifications made to Zen Cart.
	 *
	 * By default a warning will be displayed for any potential issues using
	 * Zen Cart's message stack (will appear once at the top of the admin).
	 *
	 * @param boolean $silent indicates if we should be silent or report warnings.
	 * @return boolean true if the removal is successful, false otherwise.
	 */
	protected function checkConfigure($silent = false) {
		global $messageStack;

		$success = true;
		$warnings = array();

		// First check the pre-loaded admin variables (from '/admin/includes/configure.php')
		$warnings = array_merge($warnings, $this->checkConfigureServerSetting('HTTP_CATALOG_SERVER', HTTP_CATALOG_SERVER, 'DIR_WS_CATALOG'));
		$warnings = array_merge($warnings, $this->checkConfigureServerSetting('HTTPS_CATALOG_SERVER', HTTPS_CATALOG_SERVER, 'DIR_WS_HTTPS_CATALOG'));
		$warnings = array_merge($warnings, $this->checkConfigurePathSetting('DIR_WS_CATALOG', DIR_WS_CATALOG));
		$warnings = array_merge($warnings, $this->checkConfigurePathSetting('DIR_WS_HTTPS_CATALOG', DIR_WS_HTTPS_CATALOG));

		if(count($warnings) > 0) {
			$success = false;

			// build message
			$message = sprintf(
				PLUGIN_CONFIG_ADMIN_WARNINGS,
				DIR_FS_ADMIN . DIR_WS_INCLUDES . 'configure.php'
			) . '<ul>';
			foreach($warnings as $warning) {
				$message .= '<li>' . $warning . '</li>';
			}
			$message .= '</ul>';
			$warnings = array();

			$messageStack->add($message, 'warning');
		}

		// Try to extract and check the store variables (from '/includes/configure.php')
		$store_config_file = DIR_FS_CATALOG . DIR_WS_INCLUDES . 'configure.php';
		$store_config_contents = @file_get_contents($store_config_file);
		if($store_config_contents === false) {
			$success = false;

			$messageStack->add(
				sprintf(
					PLUGIN_CONFIG_STORE_CONFIGURE,
					$store_config_file,
					'DIR_FS_CATALOG + DIR_WS_INCLUDES'
				),
				'warning'
			);
		}
		else {
			$warnings = array_merge($warnings, $this->checkConfigureServerSetting(
				'HTTP_SERVER',
				$this->getConfiguredDefine('HTTP_SERVER', $store_config_file, $store_config_contents),
				'DIR_WS_CATALOG'
			));
			$warnings = array_merge($warnings, $this->checkConfigureServerSetting(
				'HTTPS_SERVER',
				$this->getConfiguredDefine('HTTPS_SERVER', $store_config_file, $store_config_contents),
				'DIR_WS_HTTPS_CATALOG'
			));
			$warnings = array_merge($warnings, $this->checkConfigurePathSetting(
				'DIR_WS_CATALOG',
				$this->getConfiguredDefine('DIR_WS_CATALOG', $store_config_file, $store_config_contents)
			));
			$warnings = array_merge($warnings, $this->checkConfigurePathSetting(
				'DIR_WS_HTTPS_CATALOG',
				$this->getConfiguredDefine('DIR_WS_HTTPS_CATALOG', $store_config_file, $store_config_contents)
			));

			if(count($warnings) > 0) {
				$success = false;

				// build message
				$message = sprintf(
					PLUGIN_CONFIG_STORE_WARNINGS,
					DIR_FS_CATALOG . DIR_WS_INCLUDES . 'configure.php'
				) . '<ul>';
				foreach($warnings as $warning) {
					$message .= '<li>' . $warning . '</li>';
				}
				$message .= '</ul>';
				$warnings = array();

				$messageStack->add($message, 'warning');
			}
		}

		return $success;
	}

	/**
	 * Retrieves the value of a configured constant from a file without loading
	 * all the define statements found in the file. The value of the constant
	 * will be evaluated and cached in memory.
	 *
	 * @param string $key the name / key of the constant to search for.
	 * @param string $file the file to search for the constant in.
	 * @param string $content the file contents (optional).
	 * @return mixed|NULL the value of the constant or null of the constant was not found.
	 */
	protected function getConfiguredDefine($key, $file, $content = null) {
		// If we have already retrieved this key, simply return the answer.
		if(defined('TMP_' . $file . '_' . $key)) {
			return constant('TMP_' . $file . '_' . $key);
		}

		// Validate the file exists (and content is useable)
		if(!file_exists($file)) return null;
		if(!is_string($content) || $content === null) {
			$content = @file_get_contents($file);
			if($content === false) return null;
		}

		if(preg_match('|define\(\s*[\'"]' . $key . '[\'"]\s*,\s*(?!\s*\);)(.+?)\s*\);|', $content, $matches)) {
			// This code is already executing from the file when loaded
			// So using eval to load the same code is no more dangerous
			$define = 'define(\'' . 'TMP_' . $file . '_' . $key . '\',' . $matches[1] . ');';
			eval("$define");
			if(defined('TMP_' . $file . '_' . $key)) {
				return constant('TMP_' . $file . '_' . $key);
			}
		}
		return null;
	}

	/**
	 * Checks the server setting and returns a list of warnings.
	 *
	 * @param string $key the name of the setting's key
	 * @param unknown $value the value of the setting
	 * @param unknown $pathKey the name of the corresponding path's key
	 * @return multitype:string any detected warnings
	 */
	protected function checkConfigureServerSetting($key, $value, $pathKey) {
		$warning = array();

		if(preg_match('|^http://[a-z0-9\-\.]+/.+$|i', $value)) {
			$warning[] = sprintf(
				PLUGIN_CONFIG_REMOVE_TRAILING_SUBDIRECTORY,
				$key
			);
			$warning[] = sprintf(
				PLUGIN_CONFIG_SUBDIRECTORY_INFO,
				$pathKey,
				$key
			);
		}
		else if(substr($value, -1) == '/') {
			$warning[] = sprintf(
				PLUGIN_CONFIG_REMOVE_TRAILING_SLASH,
				$key
			);
		}

		return $warning;
	}

	/**
	 * Checks the path setting and returns a list of warnings.
	 *
	 * @param string $key the name of the setting's key
	 * @param unknown $value the value of the setting
	 * @return multitype:string any detected warnings
	 */
	protected function checkConfigurePathSetting($key, $value) {
		$warning = array();

		if(strlen($value) == 0) {
			$warning[] = sprintf(
				PLUGIN_CONFIG_SUBDIRECTORY_EMPTY,
				$key
			);
		}
		else {
			if(substr($value, 0, 1) != '/') {
				$warning[] = sprintf(
					PLUGIN_CONFIG_ADD_LEADING_SLASH,
					$key
				);
			}
			if(substr($value, -1) != '/') {
				$warning[] = sprintf(
					PLUGIN_CONFIG_ADD_TRAILING_SLASH,
					$key
				);
			}
		}

		return $warning;
	}

	/**
	 * Creates a generic init includes file to either install or remove this
	 * plugin. Does nothing and returns true if the file already exists.
	 *
	 * @param boolean $install true to create the install init include file,
	 *      false to create the removal init includes file.
	 * @return boolean true if the file was created successfully, false otherwise.
	 */
	protected function createInitIncludeFile($install = true) {
		global $messageStack;

		$files = $this->getAutoLoaderFiles();
		$type = ($install ? 'install' : 'remove');
		if(file_exists($files[$type])) return true;

		// Create a new installer file
		$contents = '<?php' . PHP_EOL . PHP_EOL .
			'if (!defined(\'IS_ADMIN_FLAG\')) { die(\'Illegal Access\'); }' . PHP_EOL . PHP_EOL .
			'@require_once DIR_FS_ADMIN . DIR_WS_CLASSES . \'' . strtolower($this->getUniqueKey()) . '_plugin.php\';' . PHP_EOL .
			'$plugin = new ' . strtolower($this->getUniqueKey()) . '_plugin();' . PHP_EOL .
			'$plugin->' . $type . '();' . PHP_EOL;

		if(@file_put_contents($files[$type], $contents) === false) {
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_FILE_NOT_FOUND,
					$this->getUniqueName(),
					$files[$type]
				),
				'error'
			);
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_FILE_WRITE,
					$this->getUniqueName(),
					$files[$type]
				),
				'error'
			);
			return false;
		}

		return true;
	}

	/**
	 * Creates a generic onetime auto load file to either install or remove this
	 * plugin. Does nothing and returns true if the file already exists.
	 *
	 * @param boolean $install true to create the onetime auto load install file,
	 *      false to create the onetime auto load removal file.
	 * @return boolean true if the file was created successfully, false otherwise.
	 */
	protected function createOnetimeFile($install = true) {
		global $messageStack;

		$files = $this->getAutoLoaderFiles();
		$type = ($install ? 'install' : 'remove');
		if(file_exists($files['loader'])) return true;

		// Create a new installer file
		$contents = '<?php' . PHP_EOL . PHP_EOL .
		'if (!defined(\'IS_ADMIN_FLAG\')) { die(\'Illegal Access\'); }' . PHP_EOL . PHP_EOL .
		'$autoLoadConfig[999][] = array(' . PHP_EOL . "\t" .
			'\'autoType\' => \'init_script\',' . PHP_EOL . "\t" .
  			'\'loadFile\' => \'init_' . strtolower($this->getUniqueKey()) .
  			'_' . $type . '.php\'' . PHP_EOL .
		');' . PHP_EOL;

		if(@file_put_contents($files['loader'], $contents) === false) {
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_FILE_NOT_FOUND,
					$this->getUniqueName(),
					$files['loader']
				),
				'error'
			);
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_FILE_WRITE,
					$this->getUniqueName(),
					$files['loader']
				),
				'error'
			);
			return false;
		}

		return true;
	}

	/**
	 * Replaces text in a file using a regular expression search and replace.
	 *
	 * @param string $file the file to alter
	 * @param string|array $pattern the pattern to search for. It can be either a
	 * 		string or an array with strings.
	 * @param string|array $replace the string or an array with strings to replace.
	 * @return boolean true if the update succeeds, false otherwise.
	 */
	protected function replaceInFile($file, $pattern, $replace) {
		global $messageStack;

		$success = true;

		$fileContent = @file_get_contents($file);
		if($fileContent === false) {
			$messageStack->add(
				sprintf(
					PLUGIN_INSTALL_ERROR_FILE_NOT_FOUND,
					$this->getUniqueName(),
					$file
				),
				'warning'
			);
			$success = false;
		}
		else {
			$newContent = preg_replace($pattern, $replace, $fileContent);
			if($newContent != $fileContent) {
				if(@rename($file, $file . '.bak') === false) {
					$messageStack->add(
						sprintf(
							PLUGIN_INSTALL_ERROR_FILE_BACKUP,
							$this->getUniqueName(),
							$file
						),
						'warning'
					);
				}
				if(@file_put_contents($file, $newContent) === false) {
					$messageStack->add(
						sprintf(
							PLUGIN_INSTALL_ERROR_FILE_WRITE,
							$this->getUniqueName(),
							$file
						),
						'warning'
					);
					$success = false;
					@rename($file . '.bak', $file);
				}
			}
		}

		return $success;
	}
}